(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = global || self, factory(global.GraphqlNorm = {}));
}(this, function (exports) { 'use strict';

    function getDocumentDefinitions(definitions) {
        let operationDefinition = undefined;
        const fragmentMap = {};
        for (const definition of definitions) {
            switch (definition.kind) {
                case "OperationDefinition":
                    operationDefinition = definition;
                    break;
                case "FragmentDefinition":
                    fragmentMap[definition.name.value] = definition;
                    break;
                default:
                    throw new Error("This is not an executable document");
            }
        }
        const rootFieldNode = {
            kind: "Field",
            name: {
                kind: "Name",
                value: "data"
            },
            selectionSet: {
                kind: "SelectionSet",
                selections: operationDefinition.selectionSet.selections
            }
        };
        return [fragmentMap, rootFieldNode];
    }
    function expandFragments(resolveType, obj, selectionNodes, fragmentMap) {
        const fieldNodes = [];
        for (const selectionNode of selectionNodes) {
            switch (selectionNode.kind) {
                case "Field":
                    fieldNodes.push(selectionNode);
                    break;
                case "InlineFragment": {
                    const fragmentTypeName = selectionNode.typeCondition && selectionNode.typeCondition.name.value;
                    const objTypeName = resolveType(obj);
                    // Only include this fragment if the typename matches
                    if (fragmentTypeName === objTypeName) {
                        fieldNodes.push(...expandFragments(resolveType, obj, selectionNode.selectionSet.selections, fragmentMap));
                    }
                    break;
                }
                case "FragmentSpread": {
                    const fragment = fragmentMap[selectionNode.name.value];
                    const fragmentTypeName = fragment.typeCondition && fragment.typeCondition.name.value;
                    const objTypeName = resolveType(obj);
                    // Only include this fragment if the typename matches
                    if (fragmentTypeName === objTypeName) {
                        fieldNodes.push(...expandFragments(resolveType, obj, fragment.selectionSet.selections, fragmentMap));
                    }
                    break;
                }
                default:
                    throw new Error("Unknown selection node field kind: " + selectionNode.kind);
            }
        }
        return fieldNodes;
    }
    function resolveValueNode(valueNode, variables) {
        switch (valueNode.kind) {
            case "Variable":
                return variables[valueNode.name.value];
            case "NullValue":
                return null;
            case "ListValue":
                return valueNode.values.map(f => resolveValueNode(f, variables));
            case "ObjectValue":
                const valueObject = {};
                for (const field of valueNode.fields) {
                    valueObject[field.name.value] = resolveValueNode(field.value, variables);
                }
                return valueObject;
            default:
                return valueNode.value;
        }
    }
    function fieldNameWithArguments(fieldNode, variables) {
        const argumentsObject = {};
        for (const argumentNode of fieldNode.arguments) {
            argumentsObject[argumentNode.name.value] = resolveValueNode(argumentNode.value, variables);
        }
        const hashedArgs = JSON.stringify(argumentsObject);
        return fieldNode.name.value + "(" + hashedArgs + ")";
    }
    const defaultGetObjectId = (object) => {
        return object.id === undefined
            ? undefined
            : `${object.__typename}:${object.id}`;
    };
    const defaultResolveType = (object) => {
        if (object.__typename === undefined) {
            throw new Error("__typename cannot be undefined.");
        }
        return object.__typename;
    };
    /**
     * Evaluates  @skip and @include directives on field
     * and returns true if the node should be included.
     */
    function shouldIncludeField(directives, variables = {}) {
        let finalInclude = true;
        for (const directive of directives) {
            let directiveInclude = true;
            if (directive.name.value === "skip" || directive.name.value === "include") {
                if (directive.arguments) {
                    for (const arg of directive.arguments) {
                        if (arg.name.value === "if") {
                            let argValue;
                            if (arg.value.kind === "Variable") {
                                argValue = variables && !!variables[arg.value.name.value];
                            }
                            else if (arg.value.kind === "BooleanValue") {
                                argValue = arg.value.value;
                            }
                            else {
                                // The if argument must be of type Boolean!
                                // http://facebook.github.io/graphql/June2018/#sec--skip
                                throw new Error(`The if argument must be of type Boolean!, found '${arg.value.kind}'`);
                            }
                            let argInclude = directive.name.value === "include" ? argValue : !argValue;
                            directiveInclude = directiveInclude && argInclude;
                        }
                    }
                }
                finalInclude = finalInclude && directiveInclude;
            }
        }
        return finalInclude;
    }

    /**
     * Normalizes a graphql response.
     * @param query The graphql query document
     * @param variables The graphql query variables
     * @param response The graphql response
     * @param getObjectId Function to get normalized map key from an object
     * @param resolveType Function get get typeName from an object
     */
    function normalize(query, variables, data, getObjectId = defaultGetObjectId, resolveType = defaultResolveType) {
        const [fragmentMap, rootFieldNode] = getDocumentDefinitions(query.definitions);
        const stack = [];
        const normMap = Object.create(null);
        // Seed stack with undefined parent and "fake" getObjectId
        stack.push([rootFieldNode, Object.create(null), data, "ROOT_QUERY"]);
        let getObjectIdToUse = _ => "ROOT_QUERY";
        // The stack has work items, depending on the work item we have four different cases to handle:
        // field + responseObject + parentNormObj = normalize(responseObject) => [ID, workitems] and parentNormObj[field] = ID
        // field + responseObject + parentArray  = normalize(responseObject) => [ID, workitems] and parentArray.push(ID)
        // field + responseArray  + parentNormObj = stack.push(workItemsFrom(responseArray)) and parentNormObj[field] = new Array()
        // field + responseArray  + parentArray  = stack.push(workItemsFrom(responseArray)) and parentArray.push(new Array())
        let firstIteration = true;
        while (stack.length > 0) {
            const [fieldNode, parentNormObjOrArray, responseObjectOrArray, fallbackId] = stack.pop();
            let keyOrNewParentArray = null;
            if (responseObjectOrArray === null) {
                keyOrNewParentArray = null;
            }
            else if (!Array.isArray(responseObjectOrArray)) {
                const responseObject = responseObjectOrArray;
                // console.log("responseObject", responseObject);
                const objectToIdResult = getObjectIdToUse(responseObject);
                keyOrNewParentArray = objectToIdResult ? objectToIdResult : fallbackId;
                // Get or create normalized object
                let normObj = normMap[keyOrNewParentArray];
                if (!normObj) {
                    normObj = Object.create(null);
                    normMap[keyOrNewParentArray] = normObj;
                }
                // Expand any fragments
                const expandedSelections = expandFragments(resolveType, responseObjectOrArray, fieldNode.selectionSet.selections, fragmentMap);
                // For each field in the selection-set that has a sub-selection-set we push a work item.
                // For primtivies fields we set them directly on the normalized object.
                for (const field of expandedSelections) {
                    // Check if this field should be skipped according to @skip and @include directives
                    const include = field.directives
                        ? shouldIncludeField(field.directives, variables)
                        : true;
                    if (include) {
                        const responseFieldValue = responseObject[(field.alias && field.alias.value) || field.name.value];
                        const normFieldName = field.arguments && field.arguments.length > 0
                            ? fieldNameWithArguments(field, variables)
                            : field.name.value;
                        if (responseFieldValue !== null && field.selectionSet) {
                            // Put a work-item on the stack to normalize this field and set it on the normalized object
                            stack.push([
                                field,
                                normObj,
                                responseFieldValue,
                                //path + "." + normFieldName
                                // Use the current key plus fieldname as fallback id
                                keyOrNewParentArray + "." + normFieldName
                            ]);
                        }
                        else {
                            // This field is a primitive (not a array of normalized objects or a single normalized object)
                            normObj[normFieldName] = responseFieldValue;
                        }
                    }
                }
            }
            else {
                const responseArray = responseObjectOrArray;
                keyOrNewParentArray = [];
                for (let i = 0; i < responseArray.length; i++) {
                    stack.push([
                        fieldNode,
                        keyOrNewParentArray,
                        responseArray[i],
                        fallbackId + "." + i.toString()
                    ]);
                }
            }
            // Add to the parent, either field or an array
            if (Array.isArray(parentNormObjOrArray)) {
                const parentArray = parentNormObjOrArray;
                parentArray.unshift(keyOrNewParentArray);
            }
            else {
                const key = fieldNode.arguments && fieldNode.arguments.length > 0
                    ? fieldNameWithArguments(fieldNode, variables)
                    : fieldNode.name.value;
                const parentNormObj = parentNormObjOrArray;
                parentNormObj[key] = keyOrNewParentArray;
            }
            // Use fake objectId function only for the first iteration, then switch to the real one
            if (firstIteration) {
                getObjectIdToUse = getObjectId;
                firstIteration = false;
            }
        }
        return normMap;
    }

    /**
     * Creates a graphql response by denormalizing.
     * @param query The graphql query document
     * @param variables The graphql query variables
     * @param normMap The map of normalized objects
     * @param resolveType Function get get typeName from an object
     */
    function denormalize(query, variables, normMap, resolveType = defaultResolveType) {
        const [fragmentMap, rootFieldNode] = getDocumentDefinitions(query.definitions);
        const stack = [];
        const response = {};
        let partial = false;
        const usedFieldsMap = {};
        stack.push([rootFieldNode, "ROOT_QUERY", response, undefined]);
        while (stack.length > 0) {
            const [fieldNode, idOrIdArray, parentObjectOrArray, parentResponseKey] = stack.pop();
            // The stack has work items, depending on the work item we have four different cases to handle:
            // field + id      + parentObject = denormalize(ID) => [responseObject, workitems] and parentObject[field] = responseObject
            // field + id      + parentArray  = denormalize(ID) => [responseObject, workitems] and parentArray.push(responseObject)
            // field + idArray + parentObject = stack.push(workItemsFrom(idArray)) and parentObject[field] = new Array()
            // field + idArray + parentArray  = stack.push(workItemsFrom(idArray)) and parentArray.push(new Array())
            let responseObjectOrNewParentArray;
            if (idOrIdArray === null) {
                responseObjectOrNewParentArray = null;
            }
            else if (!Array.isArray(idOrIdArray)) {
                const key = idOrIdArray;
                const normObj = normMap[key];
                // Does not exist in normalized map. We can't fully resolve query
                if (normObj === undefined) {
                    partial = true;
                    break;
                }
                let usedFields = usedFieldsMap[key];
                if (usedFields === undefined) {
                    usedFields = new Set();
                    usedFieldsMap[key] = usedFields;
                }
                // If we've been here before we need to use the previously created response object
                if (Array.isArray(parentObjectOrArray)) {
                    responseObjectOrNewParentArray =
                        parentObjectOrArray[parentResponseKey] || Object.create(null);
                }
                else {
                    responseObjectOrNewParentArray =
                        parentObjectOrArray[parentResponseKey] || Object.create(null);
                }
                // Expand fragments and loop all fields
                const expandedSelections = expandFragments(resolveType, normObj, fieldNode.selectionSet.selections, fragmentMap);
                for (const field of expandedSelections) {
                    // Check if this field should be skipped according to @skip and @include directives
                    const include = field.directives
                        ? shouldIncludeField(field.directives, variables)
                        : true;
                    if (include) {
                        // Build key according to any arguments
                        const key = field.arguments && field.arguments.length > 0
                            ? fieldNameWithArguments(field, variables)
                            : field.name.value;
                        // Add this to used fields
                        usedFields.add(key);
                        const normObjValue = normObj[key];
                        if (normObjValue !== null && field.selectionSet) {
                            // Put a work-item on the stack to build this field and set it on the response object
                            stack.push([
                                field,
                                normObjValue,
                                responseObjectOrNewParentArray,
                                (field.alias && field.alias.value) || field.name.value
                            ]);
                        }
                        else {
                            // This field is a primitive (not a array or object)
                            if (normObjValue !== undefined) {
                                responseObjectOrNewParentArray[(field.alias && field.alias.value) || field.name.value] = normObjValue;
                            }
                            else {
                                partial = true;
                            }
                        }
                    }
                }
            }
            else {
                const idArray = idOrIdArray;
                responseObjectOrNewParentArray =
                    parentObjectOrArray[parentResponseKey] || [];
                for (let i = 0; i < idArray.length; i++) {
                    const idArrayItem = idArray[i];
                    stack.push([
                        fieldNode,
                        idArrayItem,
                        responseObjectOrNewParentArray,
                        i
                    ]);
                }
            }
            // Add to the parent, either field or an array
            if (Array.isArray(parentObjectOrArray)) {
                const parentArray = parentObjectOrArray;
                parentArray[parentResponseKey] = responseObjectOrNewParentArray;
            }
            else {
                const parentObject = parentObjectOrArray;
                parentObject[parentResponseKey ||
                    (fieldNode.alias && fieldNode.alias.value) ||
                    fieldNode.name.value] = responseObjectOrNewParentArray;
            }
        }
        const data = response.data;
        return {
            data: !partial ? data : undefined,
            fields: usedFieldsMap
        };
    }

    /**
     * An optimized function to merge two maps of normalized objects (as returned from normalize)
     * @param normMap The first normalized map
     * @param newNormMap The second normalized map
     */
    function merge(normMap, newNormMap) {
        const updatedNormMap = Object.keys(newNormMap).reduce((stateSoFar, current) => {
            const newNormObj = Object.assign({}, (normMap[current] || {}), newNormMap[current]);
            stateSoFar[current] = newNormObj;
            return stateSoFar;
        }, {});
        return Object.assign({}, normMap, updatedNormMap);
    }

    /**
     * Fetches an object from the normalized map, taking as type parameter the object
     * has when denormalized, but returning the type it will have as a normalized object.
     * @param key The normalized map key
     * @param normMap The normalized map
     */
    function getNormalizedObject(key, normMap) {
        return normMap[key];
    }

    exports.defaultGetObjectId = defaultGetObjectId;
    exports.denormalize = denormalize;
    exports.getNormalizedObject = getNormalizedObject;
    exports.merge = merge;
    exports.normalize = normalize;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
