import { expandFragments, getDocumentDefinitions, fieldNameWithArguments, shouldIncludeField, defaultResolveType } from "./functions";
/**
 * Creates a graphql response by denormalizing.
 * @param query The graphql query document
 * @param variables The graphql query variables
 * @param normMap The map of normalized objects
 * @param resolveType Function get get typeName from an object
 */
export function denormalize(query, variables, normMap, resolveType = defaultResolveType) {
    const [fragmentMap, rootFieldNode] = getDocumentDefinitions(query.definitions);
    const stack = [];
    const response = {};
    let partial = false;
    const usedFieldsMap = {};
    stack.push([rootFieldNode, "ROOT_QUERY", response, undefined]);
    while (stack.length > 0) {
        const [fieldNode, idOrIdArray, parentObjectOrArray, parentResponseKey] = stack.pop();
        // The stack has work items, depending on the work item we have four different cases to handle:
        // field + id      + parentObject = denormalize(ID) => [responseObject, workitems] and parentObject[field] = responseObject
        // field + id      + parentArray  = denormalize(ID) => [responseObject, workitems] and parentArray.push(responseObject)
        // field + idArray + parentObject = stack.push(workItemsFrom(idArray)) and parentObject[field] = new Array()
        // field + idArray + parentArray  = stack.push(workItemsFrom(idArray)) and parentArray.push(new Array())
        let responseObjectOrNewParentArray;
        if (idOrIdArray === null) {
            responseObjectOrNewParentArray = null;
        }
        else if (!Array.isArray(idOrIdArray)) {
            const key = idOrIdArray;
            const normObj = normMap[key];
            // Does not exist in normalized map. We can't fully resolve query
            if (normObj === undefined) {
                partial = true;
                break;
            }
            let usedFields = usedFieldsMap[key];
            if (usedFields === undefined) {
                usedFields = new Set();
                usedFieldsMap[key] = usedFields;
            }
            // If we've been here before we need to use the previously created response object
            if (Array.isArray(parentObjectOrArray)) {
                responseObjectOrNewParentArray =
                    parentObjectOrArray[parentResponseKey] || Object.create(null);
            }
            else {
                responseObjectOrNewParentArray =
                    parentObjectOrArray[parentResponseKey] || Object.create(null);
            }
            // Expand fragments and loop all fields
            const expandedSelections = expandFragments(resolveType, normObj, fieldNode.selectionSet.selections, fragmentMap);
            for (const field of expandedSelections) {
                // Check if this field should be skipped according to @skip and @include directives
                const include = field.directives
                    ? shouldIncludeField(field.directives, variables)
                    : true;
                if (include) {
                    // Build key according to any arguments
                    const key = field.arguments && field.arguments.length > 0
                        ? fieldNameWithArguments(field, variables)
                        : field.name.value;
                    // Add this to used fields
                    usedFields.add(key);
                    const normObjValue = normObj[key];
                    if (normObjValue !== null && field.selectionSet) {
                        // Put a work-item on the stack to build this field and set it on the response object
                        stack.push([
                            field,
                            normObjValue,
                            responseObjectOrNewParentArray,
                            (field.alias && field.alias.value) || field.name.value
                        ]);
                    }
                    else {
                        // This field is a primitive (not a array or object)
                        if (normObjValue !== undefined) {
                            responseObjectOrNewParentArray[(field.alias && field.alias.value) || field.name.value] = normObjValue;
                        }
                        else {
                            partial = true;
                        }
                    }
                }
            }
        }
        else {
            const idArray = idOrIdArray;
            responseObjectOrNewParentArray =
                parentObjectOrArray[parentResponseKey] || [];
            for (let i = 0; i < idArray.length; i++) {
                const idArrayItem = idArray[i];
                stack.push([
                    fieldNode,
                    idArrayItem,
                    responseObjectOrNewParentArray,
                    i
                ]);
            }
        }
        // Add to the parent, either field or an array
        if (Array.isArray(parentObjectOrArray)) {
            const parentArray = parentObjectOrArray;
            parentArray[parentResponseKey] = responseObjectOrNewParentArray;
        }
        else {
            const parentObject = parentObjectOrArray;
            parentObject[parentResponseKey ||
                (fieldNode.alias && fieldNode.alias.value) ||
                fieldNode.name.value] = responseObjectOrNewParentArray;
        }
    }
    const data = response.data;
    return {
        data: !partial ? data : undefined,
        fields: usedFieldsMap
    };
}
//# sourceMappingURL=denormalize.js.map